/*
 * Copyright (C) 2010 Canonical, Ltd.
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * version 3.0 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License version 3.0 for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 * Authored by:
 *               Mikkel Kamstrup Erlandsen <mikkel.kamstrup@canonical.com>
 */

/**
 * SECTION:dee-versioned-model
 * @short_description: Abstract base class for easing implementations of
 *                     #DeeModel<!-- -->s providing a unique version number
 *                     for each row
 * @include: dee.h
 *
 * #DeeVersionedModel is an abstract base class to ease implementation of
 * #DeeModel<!-- -->s providing rows versioned by a
 * <emphasis>sequence number</emphasis>.
 *
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <memory.h>
#include <time.h>
#include <unistd.h>
#include <gobject/gvaluecollector.h>

#include "dee-model.h"
#include "dee-versioned-model.h"
#include "dee-marshal.h"
#include "trace-log.h"

static void dee_versioned_model_model_iface_init (DeeModelIface *iface);
G_DEFINE_ABSTRACT_TYPE_WITH_CODE (DeeVersionedModel,
                                  dee_versioned_model,
                                  G_TYPE_OBJECT,
                                  G_IMPLEMENT_INTERFACE (DEE_TYPE_MODEL,
                                                         dee_versioned_model_model_iface_init));

#define DEE_VERSIONED_MODEL_GET_PRIVATE(obj) \
  (G_TYPE_INSTANCE_GET_PRIVATE(obj, DEE_TYPE_VERSIONED_MODEL, DeeVersionedModelPrivate))

/**
 * DeeVersionedModelPrivate:
 *
 * Ignore this structure.
 */
struct _DeeVersionedModelPrivate
{
  /* Seqnum tracking */
  GArray     *seqnums;
  guint64     seqnum;

  /* Column type info */
  guint       n_columns;
  gchar     **column_schemas; // NULL terminated
  
};

/*
 * Public overridable DeeVersionedModel methods
 */
static DeeSeqnumState
                 dee_versioned_model_check_seqnums_real    (DeeModel *self,
                                                            GArray *seqnums);

static DeeSeqnumState
                 dee_versioned_model_check_seqnum_real     (DeeModel     *self,
                                                            guint64       seqnum);

static guint64   dee_versioned_model_inc_seqnum_real       (DeeModel     *self,
                                                            guint         pos);
                                                          
static void      dee_versioned_model_set_seqnum_real       (DeeModel     *self,
                                                            guint         pos,
                                                            guint64       seqnum);

static guint64   dee_versioned_model_append_next_seqnum_real  (DeeModel     *self);

static guint64   dee_versioned_model_prepend_next_seqnum_real (DeeModel     *self);

static guint64   dee_versioned_model_insert_next_seqnum_real  (DeeModel     *self,
                                                               guint         pos);

static guint64   dee_versioned_model_get_seqnum_real       (DeeModel     *self,
                                                            guint         pos);

static guint64   dee_versioned_model_get_last_seqnum_real  (DeeModel     *self);

static void      dee_versioned_model_set_last_seqnum_real  (DeeModel     *self,
                                                            guint64       seqnum);

static void      dee_versioned_model_remove_seqnum_real    (DeeModel     *self,
                                                            guint         pos);

/*
 * DeeModel forward declarations
 */
static const gchar* const*   dee_versioned_model_get_schema (DeeModel *self,
                                                             guint    *num_columns);

static const gchar*   dee_versioned_model_get_column_schema (DeeModel *self,
                                                             guint     column);

static void           dee_versioned_model_set_schema_full  (DeeModel           *self,
                                                            const gchar* const *column_schemas,
                                                            guint               num_columns);

static guint          dee_versioned_model_get_n_columns  (DeeModel *self);

static guint          dee_versioned_model_get_n_rows     (DeeModel *self);

static void           dee_versioned_model_clear          (DeeModel *self);

static DeeModelIter*  dee_versioned_model_append_row  (DeeModel  *self,
                                                       GVariant **row_members);

static DeeModelIter*  dee_versioned_model_prepend_row  (DeeModel  *self,
                                                        GVariant **row_members);

static DeeModelIter*  dee_versioned_model_insert_row  (DeeModel  *self,
                                                       guint      pos,
                                                       GVariant **row_members);

static DeeModelIter*  dee_versioned_model_insert_row_before (DeeModel      *self,
                                                             DeeModelIter  *iter,
                                                             GVariant     **row_members);

static void           dee_versioned_model_remove         (DeeModel     *self,
                                                          DeeModelIter *iter);

static void           dee_versioned_model_set_value      (DeeModel       *self,
                                                          DeeModelIter   *iter,
                                                          guint           column,
                                                          GVariant       *value);

static void           dee_versioned_model_set_row            (DeeModel       *self,
                                                              DeeModelIter   *iter,
                                                              GVariant      **row_members);

static GVariant*     dee_versioned_model_get_value      (DeeModel     *self,
                                                          DeeModelIter *iter,
                                                          guint         column);

static DeeModelIter* dee_versioned_model_get_first_iter  (DeeModel     *self);

static DeeModelIter* dee_versioned_model_get_last_iter   (DeeModel     *self);

static DeeModelIter* dee_versioned_model_get_iter_at_row (DeeModel     *self,
                                                         guint          row);

static gboolean       dee_versioned_model_get_bool       (DeeModel    *self,
                                                         DeeModelIter *iter,
                                                         guint         column);

static guchar         dee_versioned_model_get_uchar      (DeeModel     *self,
                                                         DeeModelIter *iter,
                                                         guint          column);

static gint32         dee_versioned_model_get_int32     (DeeModel     *self,
                                                         DeeModelIter *iter,
                                                         guint          column);

static guint32        dee_versioned_model_get_uint32    (DeeModel     *self,
                                                         DeeModelIter *iter,
                                                         guint          column);

static gint64         dee_versioned_model_get_int64      (DeeModel     *self,
                                                         DeeModelIter *iter,
                                                         guint          column);

static guint64        dee_versioned_model_get_uint64     (DeeModel     *self,
                                                         DeeModelIter *iter,
                                                         guint          column);

static gdouble        dee_versioned_model_get_double     (DeeModel     *self,
                                                         DeeModelIter *iter,
                                                         guint          column);

static const gchar*   dee_versioned_model_get_string     (DeeModel     *self,
                                                         DeeModelIter *iter,
                                                         guint          column);

static DeeModelIter* dee_versioned_model_next            (DeeModel     *self,
                                                         DeeModelIter *iter);

static DeeModelIter* dee_versioned_model_prev            (DeeModel     *self,
                                                         DeeModelIter *iter);

static gboolean       dee_versioned_model_is_first       (DeeModel     *self,
                                                         DeeModelIter *iter);

static gboolean       dee_versioned_model_is_last        (DeeModel     *self,
                                                         DeeModelIter *iter);

static guint          dee_versioned_model_get_position   (DeeModel     *self,
                                                          DeeModelIter *iter);

static void           dee_versioned_model_freeze_signals (DeeModel     *self);

static void           dee_versioned_model_thaw_signals   (DeeModel     *self);

/* GObject Init */
static void
dee_versioned_model_finalize (GObject *object)
{
  //DeeVersionedModelPrivate *priv = DEE_VERSIONED_MODEL (object)->priv;

  G_OBJECT_CLASS (dee_versioned_model_parent_class)->finalize (object);
}

static void
dee_versioned_model_set_property (GObject      *object,
                                 guint         id,
                                 const GValue *value,
                                 GParamSpec   *pspec)
{
  switch (id)
    {
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, id, pspec);
      break;
    }
}

static void
dee_versioned_model_get_property (GObject    *object,
                                 guint       id,
                                 GValue     *value,
                                 GParamSpec *pspec)
{
  switch (id)
    {
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, id, pspec);
      break;
    }
}

static void
dee_versioned_model_class_init (DeeVersionedModelClass *klass)
{
  GObjectClass  *obj_class = G_OBJECT_CLASS (klass);

  obj_class->finalize     = dee_versioned_model_finalize;  
  obj_class->set_property = dee_versioned_model_set_property;
  obj_class->get_property = dee_versioned_model_get_property;

  klass->check_seqnums   = dee_versioned_model_check_seqnums_real;
  klass->check_seqnum    = dee_versioned_model_check_seqnum_real;
  klass->inc_seqnum      = dee_versioned_model_inc_seqnum_real;
  klass->set_seqnum      = dee_versioned_model_set_seqnum_real;
  klass->append_next_seqnum  = dee_versioned_model_append_next_seqnum_real;
  klass->prepend_next_seqnum = dee_versioned_model_prepend_next_seqnum_real;
  klass->insert_next_seqnum  = dee_versioned_model_insert_next_seqnum_real;
  klass->get_seqnum      = dee_versioned_model_get_seqnum_real;
  klass->get_last_seqnum = dee_versioned_model_get_last_seqnum_real;
  klass->set_last_seqnum = dee_versioned_model_set_last_seqnum_real;
  klass->remove_seqnum   = dee_versioned_model_remove_seqnum_real;

  /* Add private data */
  g_type_class_add_private (obj_class, sizeof (DeeVersionedModelPrivate));
}

static void
dee_versioned_model_init (DeeVersionedModel *model)
{
  DeeVersionedModelPrivate *priv;

  priv = model->priv = DEE_VERSIONED_MODEL_GET_PRIVATE (model);

  priv->seqnums = g_array_new (FALSE, FALSE, sizeof (guint64));
  priv->seqnum = 0;

  priv->n_columns = 0;
  priv->column_schemas = NULL;
  
  
}

static DeeSeqnumState
dee_versioned_model_check_seqnums_real (DeeModel *self, GArray *seqnums)
{
  DeeVersionedModelPrivate *priv;
  guint64                  seqnum;
  gint                     i;

  g_return_val_if_fail (DEE_IS_VERSIONED_MODEL (self), FALSE);
  g_return_val_if_fail (seqnums != NULL, FALSE);
  g_return_val_if_fail (seqnums->len > 0, FALSE);

  priv = DEE_VERSIONED_MODEL (self)->priv;

  /* Check that the incoming seqnums are consecutive */
  seqnum = g_array_index (seqnums, guint64, 0);
  for (i = 1; i < seqnums->len; i++)
    {
      if (seqnum != g_array_index (seqnums, guint64, i) - 1)
        { 
          return DEE_SEQNUM_STATE_NOT_CONSECUTIVE;
        }
      seqnum = g_array_index (seqnums, guint64, i);
    }

  /* Check how the incoming seqnums align with our own */
  seqnum = g_array_index (seqnums, guint64, 0);
  if (seqnum > priv->seqnum + 1)
    {
      return DEE_SEQNUM_STATE_IN_THE_FUTURE;
    }
  else if (seqnum < priv->seqnum + 1)
    {
      return DEE_SEQNUM_STATE_IN_THE_PAST;
    }
  
  return DEE_SEQNUM_STATE_OK;
}

/**
 * dee_versioned_model_check_seqnums:
 * @self:
 * @seqnums:
 *
 * Assert that @seqnums are consecutive and that the beggining sequence number
 * in @seqnums is exactly one higher than the last seen seqnum of this model.
 */
DeeSeqnumState
dee_versioned_model_check_seqnums (DeeModel *self, GArray *seqnums)
{
  g_return_val_if_fail (DEE_IS_VERSIONED_MODEL (self), DEE_SEQNUM_STATE_OK);

  return DEE_VERSIONED_MODEL_GET_CLASS (self)->check_seqnums (self, seqnums);
}

static DeeSeqnumState
dee_versioned_model_check_seqnum_real (DeeModel     *self,
                                      guint64       seqnum)
{
  DeeVersionedModelPrivate *priv;
  
  g_return_val_if_fail (DEE_IS_VERSIONED_MODEL (self), DEE_SEQNUM_STATE_OK);

  priv = DEE_VERSIONED_MODEL (self)->priv;

  if (priv->seqnum > seqnum)
    return DEE_SEQNUM_STATE_IN_THE_PAST;
  if (priv->seqnum < seqnum)
    return DEE_SEQNUM_STATE_IN_THE_FUTURE;
  else
    return DEE_SEQNUM_STATE_OK;
}

DeeSeqnumState
dee_versioned_model_check_seqnum (DeeModel     *self,
                                 guint64       seqnum)
{
  g_return_val_if_fail (DEE_IS_VERSIONED_MODEL (self), DEE_SEQNUM_STATE_OK);

  return DEE_VERSIONED_MODEL_GET_CLASS (self)->check_seqnum (self, seqnum);
}

static guint64
dee_versioned_model_inc_seqnum_real (DeeModel     *self,
                                    guint         pos)
{
  DeeVersionedModelPrivate *priv;
  
  g_return_val_if_fail (DEE_IS_VERSIONED_MODEL (self), 0);

  priv = DEE_VERSIONED_MODEL (self)->priv;
  g_return_val_if_fail (pos < priv->seqnums->len, 0);
  
  priv->seqnum = priv->seqnum + 1;
  g_array_index (priv->seqnums, guint64, pos) = priv->seqnum;
  
  return priv->seqnum;
}

guint64
dee_versioned_model_inc_seqnum (DeeModel     *self,
                               guint         pos)
{
  g_return_val_if_fail (DEE_IS_VERSIONED_MODEL (self), 0);

  return DEE_VERSIONED_MODEL_GET_CLASS (self)->inc_seqnum (self, pos);
}

static void
dee_versioned_model_set_seqnum_real (DeeModel     *self,
                                    guint         pos,
                                    guint64       seqnum)
{
  DeeVersionedModelPrivate *priv;
  
  g_return_if_fail (DEE_IS_VERSIONED_MODEL (self));

  priv = DEE_VERSIONED_MODEL (self)->priv;  

  g_array_index (priv->seqnums, guint64, pos) = seqnum;
}
                                                          
void
dee_versioned_model_set_seqnum (DeeModel     *self,
                               guint         pos,
                               guint64       seqnum)
{
  g_return_if_fail (DEE_IS_VERSIONED_MODEL (self));

  DEE_VERSIONED_MODEL_GET_CLASS (self)->set_seqnum (self, pos, seqnum);
}

guint64
dee_versioned_model_append_next_seqnum (DeeModel     *self)
{
  g_return_val_if_fail (DEE_IS_VERSIONED_MODEL (self), 0);

  return DEE_VERSIONED_MODEL_GET_CLASS (self)->append_next_seqnum (self);
}

static guint64
dee_versioned_model_append_next_seqnum_real (DeeModel     *self)
{
  DeeVersionedModelPrivate *priv;
  
  g_return_val_if_fail (DEE_IS_VERSIONED_MODEL (self), 0);

  priv = DEE_VERSIONED_MODEL (self)->priv;
  priv->seqnum = priv->seqnum + 1;
  g_array_append_val (priv->seqnums, (priv->seqnum));
  
  return priv->seqnum;
}

guint64
dee_versioned_model_prepend_next_seqnum (DeeModel     *self)
{
  g_return_val_if_fail (DEE_IS_VERSIONED_MODEL (self), 0);

  return DEE_VERSIONED_MODEL_GET_CLASS (self)->prepend_next_seqnum (self);
}

static guint64
dee_versioned_model_prepend_next_seqnum_real (DeeModel     *self)
{
  DeeVersionedModelPrivate *priv;
  
  g_return_val_if_fail (DEE_IS_VERSIONED_MODEL (self), 0);

  priv = DEE_VERSIONED_MODEL (self)->priv;
  priv->seqnum = priv->seqnum + 1;
  g_array_prepend_val (priv->seqnums, priv->seqnum);
  
  return priv->seqnum;
}

guint64
dee_versioned_model_insert_next_seqnum (DeeModel     *self,
                                       guint         pos)
{
  g_return_val_if_fail (DEE_IS_VERSIONED_MODEL (self), 0);

  return DEE_VERSIONED_MODEL_GET_CLASS (self)->insert_next_seqnum (self, pos);
}


static guint64
dee_versioned_model_insert_next_seqnum_real (DeeModel *self, guint pos)
{
  DeeVersionedModelPrivate *priv;
  
  g_return_val_if_fail (DEE_IS_VERSIONED_MODEL (self), 0);
  g_return_val_if_fail (pos >= 0, 0);

  priv = DEE_VERSIONED_MODEL (self)->priv;
  g_return_val_if_fail (pos < priv->seqnums->len, 0);
  
  priv->seqnum = priv->seqnum + 1;
  g_array_insert_val (priv->seqnums, pos, priv->seqnum);
  
  return priv->seqnum;
}

static guint64
dee_versioned_model_get_seqnum_real (DeeModel     *self,
                                    guint         pos)
{
  DeeVersionedModelPrivate *priv;
  
  g_return_val_if_fail (DEE_IS_VERSIONED_MODEL (self), 0);

  priv = DEE_VERSIONED_MODEL (self)->priv;  
  g_return_val_if_fail (pos < priv->seqnums->len, 0);

  return g_array_index (priv->seqnums, guint64, pos);
}
                                                          
guint64
dee_versioned_model_get_seqnum (DeeModel     *self,
                               guint         pos)
{
  g_return_val_if_fail (DEE_IS_VERSIONED_MODEL (self), 0);

  return DEE_VERSIONED_MODEL_GET_CLASS (self)->get_seqnum (self, pos);
}

static guint64
dee_versioned_model_get_last_seqnum_real (DeeModel *self)
{
  g_return_val_if_fail (DEE_IS_VERSIONED_MODEL (self), 0);
  
  return DEE_VERSIONED_MODEL (self)->priv->seqnum;
}
                                                          
guint64
dee_versioned_model_get_last_seqnum (DeeModel *self)
{
  g_return_val_if_fail (DEE_IS_VERSIONED_MODEL (self), 0);

  return DEE_VERSIONED_MODEL_GET_CLASS (self)->get_last_seqnum (self);
}

static void
dee_versioned_model_set_last_seqnum_real (DeeModel *self,
                                          guint64   seqnum)
{
  g_return_if_fail (DEE_IS_VERSIONED_MODEL (self));
    
  DEE_VERSIONED_MODEL (self)->priv->seqnum = seqnum;
}
                                                          
void
dee_versioned_model_set_last_seqnum (DeeModel *self,
                                     guint64   seqnum)
{
  g_return_if_fail (DEE_IS_VERSIONED_MODEL (self));

  DEE_VERSIONED_MODEL_GET_CLASS (self)->set_last_seqnum (self, seqnum);
}

static void
dee_versioned_model_remove_seqnum_real (DeeModel *self,
                                       guint       pos)
{
  DeeVersionedModelPrivate *priv;
  
  g_return_if_fail (DEE_IS_VERSIONED_MODEL (self));

  priv = DEE_VERSIONED_MODEL (self)->priv;  
  g_return_if_fail (pos < priv->seqnums->len);
  
  g_array_remove_index (priv->seqnums, pos);
}


void
dee_versioned_model_remove_seqnum (DeeModel *self,
                                  guint       pos)
{
  g_return_if_fail (DEE_IS_VERSIONED_MODEL (self));

  DEE_VERSIONED_MODEL_GET_CLASS (self)->remove_seqnum (self, pos);
}


/*
 * DeeModel API
 */

static void
dee_versioned_model_set_schema_full  (DeeModel           *self,
                                      const gchar* const *column_schemas,
                                      guint               num_columns)
{
  DeeVersionedModelPrivate  *priv;
  gchar                    **column_schemas_copy;
  guint                      i;
  
  g_return_if_fail (DEE_IS_VERSIONED_MODEL (self));
  g_return_if_fail (column_schemas != NULL);

  priv = DEE_VERSIONED_MODEL (self)->priv;

  if (priv->column_schemas != NULL)
    {
      g_critical ("The DeeModel %p already has a schema", self);
      return;
    }

  /* Allocate space to store the column schemas. We NULL terminate it
   * in order to play well with g_strfreev() */
  column_schemas_copy = g_new0 (gchar*, num_columns + 1);

  /* Validate the type strings and copy the schema for our selves */
  for (i = 0; i < num_columns; i++)
    {
      if (!g_variant_type_string_is_valid (column_schemas[i]))
        {
          g_critical ("When setting schema for DeeModel %p: '%s' is not a "
                      "valid type string", self, column_schemas[i]);
          return;
        }
      column_schemas_copy[i] = g_strdup (column_schemas[i]);
    }

  /* Register the validated types as our column types */
  priv->column_schemas = column_schemas_copy; // steal
  priv->n_columns = num_columns;

#ifdef ENABLE_TRACE_LOG
  gchar* schema = g_strjoinv (", ", priv->column_schemas);
  trace_object (self, "Set schema: (%s)", schema);
  g_free (schema);
#endif
}

static const gchar* const*
dee_versioned_model_get_schema (DeeModel *self,
                                guint    *n_columns)
{
  DeeVersionedModelPrivate *priv;

  g_return_val_if_fail (DEE_IS_VERSIONED_MODEL (self), NULL);

  priv = DEE_VERSIONED_MODEL (self)->priv;

  if (n_columns != NULL)
    *n_columns = priv->n_columns;

  return (const gchar**) priv->column_schemas;
}

static const gchar*
dee_versioned_model_get_column_schema (DeeModel *self,
                                       guint     column)
{
  DeeVersionedModelPrivate *priv;
  
  g_return_val_if_fail (DEE_IS_VERSIONED_MODEL (self), NULL);

  priv = DEE_VERSIONED_MODEL (self)->priv;
  g_return_val_if_fail (column < priv->n_columns, NULL);

  return priv->column_schemas[column];
}

static guint
dee_versioned_model_get_n_columns (DeeModel *self)
{
  g_return_val_if_fail (DEE_IS_VERSIONED_MODEL (self), 0);

  return DEE_VERSIONED_MODEL (self)->priv->n_columns;
}

static guint
dee_versioned_model_get_n_rows (DeeModel *self)
{
  DeeModelIter *iter;
  guint         count;

  count = 0;
  iter = dee_model_get_first_iter (self);
  while (!dee_model_is_last (self, iter))
    {
      iter = dee_model_next (self, iter);
      count++;
    }

  return count;
}

static void
dee_versioned_model_clear (DeeModel *self)
{
  DeeModelIter            *iter;

  g_return_if_fail (DEE_IS_VERSIONED_MODEL (self));

  iter = dee_model_get_first_iter (self);

  while (!dee_model_is_last (self, iter))
    {
      dee_model_remove (self, iter);

      iter = dee_model_get_first_iter (self);      
    }
}

static DeeModelIter*
dee_versioned_model_prepend_row (DeeModel  *self,
                                 GVariant **row_members)
{
  g_critical ("%s not implemented", G_STRFUNC);
  return NULL;
}

static DeeModelIter*
dee_versioned_model_append_row (DeeModel  *self,
                                GVariant **row_members)
{
  g_critical ("%s not implemented", G_STRFUNC);
  return NULL;
}

static DeeModelIter*
dee_versioned_model_insert_row (DeeModel  *self,
                                guint      pos,
                                GVariant **row_members)
{
  DeeModelIter *iter;

  g_return_val_if_fail (DEE_IS_VERSIONED_MODEL (self), 0);

  if (pos > 0)
    {
      iter = dee_model_get_iter_at_row (self, pos);
      return dee_model_insert_row_before (self, iter, row_members);
    }
  else if (pos == 0)
    return dee_model_prepend_row (self, row_members);
  else
    return dee_model_append_row (self, row_members);
}

static DeeModelIter*
dee_versioned_model_insert_row_before (DeeModel      *self,
                                       DeeModelIter  *iter,
                                       GVariant     **row_members)
{
  g_critical ("%s not implemented", G_STRFUNC);
  return NULL;
}

static void
dee_versioned_model_remove (DeeModel     *self,
                            DeeModelIter *iter_)
{
  g_critical ("%s not implemented", G_STRFUNC);
}

static void
dee_versioned_model_set_value (DeeModel       *self,
                               DeeModelIter   *iter,
                               guint           column,
                               GVariant       *value)
{
  g_critical ("%s not implemented", G_STRFUNC);
}

static void
dee_versioned_model_set_row (DeeModel       *self,
                             DeeModelIter   *iter,
                             GVariant      **row_members)
{
  g_critical ("%s not implemented", G_STRFUNC);
}

static GVariant*
dee_versioned_model_get_value (DeeModel     *self,
                               DeeModelIter *iter,
                               guint         column)
{
  g_critical ("%s not implemented", G_STRFUNC);

  return NULL;
}

static gboolean
dee_versioned_model_get_bool (DeeModel     *self,
                              DeeModelIter *iter,
                              guint         column)
{
  GVariant *value;
  gboolean  b;

  g_return_val_if_fail (DEE_IS_VERSIONED_MODEL (self), FALSE);

  value = dee_model_get_value (self, iter, column);
  b = g_variant_get_boolean (value);
  g_variant_unref (value);

  return b;
}

static guchar
dee_versioned_model_get_uchar (DeeModel      *self,
                               DeeModelIter  *iter,
                               guint          column)
{
  GVariant *value;
  guchar    u;

  g_return_val_if_fail (DEE_IS_VERSIONED_MODEL (self), FALSE);

  value = dee_model_get_value (self, iter, column);
  u = g_variant_get_byte(value);
  g_variant_unref (value);

  return u;
}

static gint32
dee_versioned_model_get_int32 (DeeModel        *self,
                               DeeModelIter    *iter,
                               guint            column)
{
  GVariant *value;
  gint32    i;

  g_return_val_if_fail (DEE_IS_VERSIONED_MODEL (self), FALSE);

  value = dee_model_get_value (self, iter, column);
  i = g_variant_get_int32 (value);
  g_variant_unref (value);

  return i;
}

static guint32
dee_versioned_model_get_uint32 (DeeModel      *self,
                                DeeModelIter  *iter,
                                guint          column)
{
  GVariant *value;
  guint32    u;

  g_return_val_if_fail (DEE_IS_VERSIONED_MODEL (self), FALSE);

  value = dee_model_get_value (self, iter, column);
  u = g_variant_get_uint32 (value);
  g_variant_unref (value);

  return u;
}


static gint64
dee_versioned_model_get_int64 (DeeModel      *self,
                               DeeModelIter  *iter,
                               guint          column)
{
  GVariant *value;
  gint64    i;

  g_return_val_if_fail (DEE_IS_VERSIONED_MODEL (self), FALSE);

  value = dee_model_get_value (self, iter, column);
  i = g_variant_get_int64 (value);
  g_variant_unref (value);

  return i;
}


static guint64
dee_versioned_model_get_uint64 (DeeModel      *self,
                                DeeModelIter  *iter,
                                guint          column)
{
  GVariant *value;
  guint64   u;

  g_return_val_if_fail (DEE_IS_VERSIONED_MODEL (self), FALSE);

  value = dee_model_get_value (self, iter, column);
  u = g_variant_get_uint64 (value);
  g_variant_unref (value);

  return u;
}

static gdouble
dee_versioned_model_get_double (DeeModel      *self,
                                DeeModelIter  *iter,
                                guint          column)
{
  GVariant *value;
  gdouble   d;

  g_return_val_if_fail (DEE_IS_VERSIONED_MODEL (self), FALSE);

  value = dee_model_get_value (self, iter, column);
  d = g_variant_get_double (value);
  g_variant_unref (value);

  return d;
}

static const gchar*
dee_versioned_model_get_string (DeeModel      *self,
                                DeeModelIter  *iter,
                                guint          column)
{
  GVariant    *value;
  const gchar *s;

  g_return_val_if_fail (DEE_IS_VERSIONED_MODEL (self), FALSE);

  value = dee_model_get_value (self, iter, column);
  s = g_variant_get_string (value, NULL);
  g_variant_unref (value);

  return s;
}

static DeeModelIter*
dee_versioned_model_get_first_iter (DeeModel     *self)
{
  g_critical ("%s not implemented", G_STRFUNC);
  return NULL;
}

static DeeModelIter*
dee_versioned_model_get_last_iter (DeeModel *self)
{
  DeeModelIter *iter;

  g_return_val_if_fail (DEE_IS_VERSIONED_MODEL (self), NULL);

  iter = dee_model_get_first_iter (self);
  while (!dee_model_is_last (self, iter))
    iter = dee_model_next (self, iter);

  return iter;
}

static DeeModelIter*
dee_versioned_model_get_iter_at_row (DeeModel *self,
                                     guint     row)
{
  DeeModelIter *iter;
  guint         pos;

  g_return_val_if_fail (DEE_IS_VERSIONED_MODEL (self), NULL);

  pos = 0;
  iter = dee_model_get_first_iter (self);
  while (!dee_model_is_last (self, iter) && pos < row)
    iter = dee_model_next (self, iter);
    pos++;

  if (dee_model_is_last (self, iter))
    {
      g_critical ("Index %u is out of bounds in model of size %u",
                  row, pos);
    }

  return iter;
}

static DeeModelIter*
dee_versioned_model_next (DeeModel     *self,
                          DeeModelIter *iter)
{
  g_critical ("%s not implemented", G_STRFUNC);
  return NULL;
}

static DeeModelIter*
dee_versioned_model_prev (DeeModel     *self,
                          DeeModelIter *iter)
{
  g_critical ("%s not implemented", G_STRFUNC);
  return NULL;
}

static gboolean
dee_versioned_model_is_first (DeeModel     *self,
                              DeeModelIter *iter)
{
  DeeModelIter *first;
  
  g_return_val_if_fail (DEE_IS_VERSIONED_MODEL (self), FALSE);

  first = dee_model_get_first_iter (self);
  return first == iter;
}

static gboolean
dee_versioned_model_is_last (DeeModel     *self,
                             DeeModelIter *iter)
{
  DeeModelIter *last;
  
  g_return_val_if_fail (DEE_IS_VERSIONED_MODEL (self), FALSE);

  last = dee_model_get_last_iter (self);
  return last == iter;
}

static guint
dee_versioned_model_get_position (DeeModel     *self,
                                  DeeModelIter *iter)
{
  DeeModelIter *_iter;
  guint          pos;

  g_return_val_if_fail (DEE_IS_VERSIONED_MODEL (self), 0);

  pos = 0;
  _iter = dee_model_get_first_iter (self);
  while (!dee_model_is_last (self, iter) && iter != _iter)
    _iter = dee_model_next (self, _iter);
    pos++;

  if (iter == _iter)
    return pos;
  else
    {
      g_critical ("Can not find position of unknown iter %p", iter);
      return -1;
    }
}

static void
dee_versioned_model_freeze_signals (DeeModel     *self)
{
  g_critical ("%s not implemented", G_STRFUNC);
}

static void
dee_versioned_model_thaw_signals (DeeModel     *self)
{
  g_critical ("%s not implemented", G_STRFUNC);
}

static void
dee_versioned_model_model_iface_init (DeeModelIface *iface)
{
  iface->set_schema_full      = dee_versioned_model_set_schema_full;
  iface->get_schema           = dee_versioned_model_get_schema;
  iface->get_column_schema    = dee_versioned_model_get_column_schema;
  iface->get_n_columns        = dee_versioned_model_get_n_columns;
  iface->get_n_rows           = dee_versioned_model_get_n_rows;
  iface->append_row           = dee_versioned_model_append_row;
  iface->prepend_row          = dee_versioned_model_prepend_row;
  iface->insert_row           = dee_versioned_model_insert_row;
  iface->insert_row_before    = dee_versioned_model_insert_row_before;
  iface->remove               = dee_versioned_model_remove;
  iface->clear                = dee_versioned_model_clear;
  iface->set_value            = dee_versioned_model_set_value;
  iface->set_row              = dee_versioned_model_set_row;
  iface->get_value            = dee_versioned_model_get_value;
  iface->get_first_iter       = dee_versioned_model_get_first_iter;
  iface->get_last_iter        = dee_versioned_model_get_last_iter;
  iface->get_iter_at_row      = dee_versioned_model_get_iter_at_row;
  iface->get_bool             = dee_versioned_model_get_bool;
  iface->get_uchar            = dee_versioned_model_get_uchar;
  iface->get_int32            = dee_versioned_model_get_int32;
  iface->get_uint32           = dee_versioned_model_get_uint32;
  iface->get_int64            = dee_versioned_model_get_int64;
  iface->get_uint64           = dee_versioned_model_get_uint64;
  iface->get_double           = dee_versioned_model_get_double;
  iface->get_string           = dee_versioned_model_get_string;
  iface->next                 = dee_versioned_model_next;
  iface->prev                 = dee_versioned_model_prev;
  iface->is_first             = dee_versioned_model_is_first;
  iface->is_last              = dee_versioned_model_is_last;
  iface->get_position         = dee_versioned_model_get_position;
  iface->freeze_signals       = dee_versioned_model_freeze_signals;
  iface->thaw_signals         = dee_versioned_model_thaw_signals;
}
